import path from 'path'
import os from 'os'
import fs from 'fs-extra'
import crypto from 'crypto'
import Handlebars from 'handlebars'
// @ts-ignore untyped metro-cache
import { FileStore, AutoCleanFileStore } from 'metro-cache';
// @ts-ignore
import { Terminal } from 'metro-core';
// @ts-ignore
import getMaxWorkers from '@hyext/matrix/src/lib/getMaxWorkers';
// @ts-ignore
import TerminalReporter from '../lib/HyTerminalReporter';

import { MetroConfig } from './configTypes'

import blacklist from './blacklist'
import { defaultFactory, createModuleIdHOF } from './createModuleIdHOF'
import resolveReactNativePath from './resolveReactNativePath'
import filterModuleHOF from './filterModuleHOF'

const DEFAULT_METRO_MINIFIER_PATH = 'metro-minify-uglify';

const INTERNAL_CALLSITES_REGEX = new RegExp(
  [
    '/Libraries/Renderer/implementations/.+\\.js$',
    '/Libraries/BatchedBridge/MessageQueue\\.js$',
    '/Libraries/YellowBox/.+\\.js$',
    '/Libraries/LogBox/.+\\.js$',
    '/Libraries/Core/Timers/.+\\.js$',
    '/node_modules/react-devtools-core/.+\\.js$',
    '/node_modules/react-refresh/.+\\.js$',
    '/node_modules/scheduler/.+\\.js$',
  ].join('|'),
);

interface BaseOptions {
  projectRoot: string
  extUuid?: string
  cacheVersion?: string
  port?: number
}

interface DevOptions extends BaseOptions {
  mode: 'dev',
  imageBaseURL?: string
  SDKAlias?: { [aliasName: string]: string[] }
  filterModules?: Set<string>
}

interface ProdOptions extends BaseOptions {
  mode: 'production'

  /*
   * 小程序打包时, 会把相对路径的图片引用改成对cdn图片的引用
   * 1. 如果打包时把图片保存到 outputDir, 那么 imageBaseURL 和 publicPath 相同
   * 2. 如果打包时把图片保存到 outputDir/app_panel/, 那么 imageBaseURL 和 publicPath/app_panel/ 相同
   *
   * 默认情况下, outputDir 是 publicPath. 除非在构建时 BundleConfig 显示指定了 assetsDir
   * 默认情况下, 图片资源会打包到跟 jsBundle 同级目录, 所以 imageBaseURL 一般是 publicPath/jsBundle_dir/
   *
   */
  imageBaseURL: string

  /*
   * SDK 的npm包名称, 包含 org 前缀(如有), 如 ['@hyext/hyext-rn-sdk'], ['hyext-rn-sdk']
   * 业务代码对 sdk 的引用将会被替换成 open-sdk(open-sdk 在app内预置了, 可直接使用)
   * 如 node_modules_@hyext_hyext-rn-sdk 将会变替换成 open-sdk
   */
  SDKAlias: { [aliasName: string]: string[] }

  filterModules: Set<string>
}


/*
 * SDKRole: 'streamer' | 'audience' 仅在线上打包使用:
 * 1. filterModuleHOF 用来过滤filterModules 里指定的包
 * 2. transform 用来替换sdk
 */
export default function getDefaultConfig(options: DevOptions | ProdOptions): MetroConfig {
  const {
    projectRoot,
    imageBaseURL = '',
    filterModules = new Set(),
    SDKAlias = {},
    mode = 'dev',
    extUuid = 'huya-ext-shared-uuid',
    cacheVersion = '1.0',
    port = 8081,
  } = options

  const defaultConfig = {
    resolver: {
      assetExts: [
        'bmp', 'gif', 'jpg', 'jpeg',
        'png', 'psd', 'svg', 'webp',
        'm4v', 'mov', 'mp4', 'mpeg',
        'mpg', 'webm', 'aac', 'aiff',
        'caf', 'm4a', 'mp3', 'wav',
        'html', 'json', 'pdf', 'yaml',
        'yml', 'otf', 'ttf', 'zip'
      ],
      sourceExts: ['js', 'json', 'ts', 'tsx', 'hycss'],
      resolverMainFields: ['react-native', 'browser', 'main'],
      platforms: ['ios', 'android', 'native', 'web'],
      extraNodeModules: {},
      resolveRequest: null,
      hasteImplModulePath: undefined,
      blacklistRE: blacklist(),
      useWatchman: true,

      virtualMapper: (file: any) => [file],
    },

    serializer: {
      polyfillModuleNames: [],

      getRunModuleStatement: (moduleId: number | string) => `__r(${JSON.stringify(moduleId)});`,

      postProcessBundleSourcemap: ({ code, map, outFileName }: any) => ({ code, map }),

      // copy from @react-native-community/cli
      getModulesRunBeforeMainModule: (entryRelative: string) => {
        return [
          require.resolve(
            path.join(resolveReactNativePath(projectRoot), 'Libraries/Core/InitializeCore'),
          ),
          // 不在项目根目录下不生效
          // path.resolve(__dirname, './runBeforeMainModule.js'),
        ]
      },

      getPolyfills: () => {
        // getPolyfills 返回一个数组
        const rnPolyfills = require(path.join(resolveReactNativePath(projectRoot), 'rn-get-polyfills'))()

        return rnPolyfills
      },

      processModuleFilter: filterModuleHOF({ projectRoot, imageBaseURL, SDKAlias, filterModules, mode }),
      createModuleIdFactory: mode === 'dev' ? defaultFactory : createModuleIdHOF({ projectRoot, SDKAlias, }),
      experimentalSerializerHook: () => { },
      customSerializer: null,
    },

    server: {
      useGlobalHotkey: true,
      port,
      enhanceMiddleware: (middleware: any, server: any, onRequest: any) => {
        return async (req: any, res: any, next: any) => {
          if (typeof onRequest === 'function') {
            await onRequest(req.url)
          }

          return middleware(req, res, next)
        }
      },
      runInspectorProxy: true,
      verifyConnections: false,
      rewriteRequestUrl: (url: string) => url,
    },

    symbolicator: {
      // copy from @react-native-community/cli
      customizeFrame: (frame: any) => {
        const collapse = Boolean(
          frame.file && INTERNAL_CALLSITES_REGEX.test(frame.file),
        );
        return { collapse };
      },
    },

    transformer: {
      assetPlugins: [],


      asyncRequireModulePath: require.resolve(
        '/Users/simon/Developer/huya/hyext-sdk-demo/asyncRequire'
      ),
      // asyncRequireModulePath: require.resolve(
      //   'metro-runtime/src/modules/asyncRequire',
      // ),

      // FIXME: 构造一个transformer 中间件, 让transformers 依次执行
      babelTransformerPath: path.resolve(__dirname, './cssTransformer.js'),

      babelPlugins: [],

      assetRegistryPath: 'react-native/Libraries/Image/AssetRegistry', // copy from @react-native-community/cli

      dynamicDepsInPackages: 'throwAtRuntime',
      enableBabelRCLookup: false,
      enableBabelRuntime: true,
      experimentalImportBundleSupport: false,
      getTransformOptions: async () => ({
        transform: {
          experimentalImportSupport: false,

          /*
           * 开发环境下, 这个值设置为true, 否则会遇到问题:
           * https://github.com/facebook/react-native/issues/28801
           */
          inlineRequires: mode === 'dev',
          unstable_disableES6Transforms: false,
        },
        preloadedModules: false,
        ramGroups: [],
      }),
      minifierConfig: {
        mangle: {
          toplevel: false,
        },
        output: {
          ascii_only: true,
          quote_style: 3,
          wrap_iife: true,
        },
        sourceMap: {
          includeSources: false,
        },
        toplevel: false,
        compress: {
          // reduce_funcs inlines single-use functions, which cause perf regressions.
          reduce_funcs: false,
        },
      },
      minifierPath: DEFAULT_METRO_MINIFIER_PATH,
      optimizationSizeLimit: 150 * 1024, // 150 KiB.
      postMinifyProcess: (x: any) => x,
      transformVariants: { default: {} },
      // workerPath: '@hyext/matrix/src/DeltaBundler/Worker',
      // workerPath: require.resolve('@hyext/matrix/src/DeltaBundler/Worker'),

      // 这个地址不要轻易改动, 否则开发环境加载不出图片
      publicPath: mode === 'dev' ? '' : '/hyext-assets',
    },

    extUuid,

    cacheStores: [
      new FileStore({
        // 使用系统的临时文件夹, 由操作系统决定何时清除
        root: path.join(
          process.env.NODE_PACK_CACHE_ROOT || os.tmpdir(),
          'huya-ext-metro-cache',
          mode === 'dev'
            ? `${path.basename(projectRoot)}@${cacheVersion}`
            : `${extUuid}@${cacheVersion}`,
        ),
      })
    ],

    // 确保不同的小程序能产生不同的缓存值
    cacheVersion: mode === 'dev'
      ? `${path.basename(projectRoot)}@${cacheVersion}`
      : `${extUuid}@${cacheVersion}`,
    // We assume the default project path is two levels up from
    // node_modules/metro/
    projectRoot,
    stickyWorkers: true,
    watchFolders: [projectRoot],
    // transformerPath: require.resolve('@hyext/matrix/src/JSTransformer/worker.js'),
    maxWorkers: getMaxWorkers(),
    resetCache: false,
    reporter: new TerminalReporter(new Terminal(process.stdout)),

    onReady: () => { },
    enableRootBabelRC: false,
  }

  return defaultConfig
}

export function generateTransformerFile(options: {
  projectRoot: string
  extUuid?: string
  cacheVersion?: string
  babelPlugins?: any[]
  hyextTransformerConfig?: any
}) {
  const {
    projectRoot,
    extUuid = 'huya-ext-shared-uuid',
    cacheVersion = '1.0',
    babelPlugins = [],
    hyextTransformerConfig = {},
  } = options

  const identity = crypto
    .createHash('sha1')
    .update([extUuid || projectRoot, cacheVersion].join('$'))
    .digest('hex')

  const tempalteFile = path.join(__dirname, './babelTransformer.js.hbs')
  const destFile = path.join(__dirname, `${identity}.babelTransformer.js`)

  const template = Handlebars.compile(fs.readFileSync(tempalteFile, 'utf8'))
  fs.writeFileSync(
    destFile,
    template({
      HANDLE_BARS_PLUGINS: JSON.stringify(babelPlugins),
      HANDLE_BARS_HYEXT_TRANSFORMER_CONFIG: JSON.stringify(hyextTransformerConfig),
    }),
    {
      encoding: 'utf8',
    },
  )

  return destFile
}
